@@ -1,9 +1,18 @@
 # cython: auto_cpdef=True
+import hashlib
+from io import StringIO
 import math
 from os import path
 from copy import deepcopy
-import json
+import re
+from typing import Tuple, Set, Optional, List
 
+from .types import DictSchema, Schema, NamedSchemas
+from .repository import (
+    FlatDictRepository,
+    SchemaRepositoryError,
+    AbstractSchemaRepository,
+)
 from ._schema_common import (
     PRIMITIVES,
     UnknownType,
@@ -11,8 +20,14 @@ from ._schema_common import (
     RESERVED_PROPERTIES,
     OPTIONAL_FIELD_PROPERTIES,
     RESERVED_FIELD_PROPERTIES,
+    JAVA_FINGERPRINT_MAPPING,
+    FINGERPRINT_ALGORITHMS,
+    RABIN_64,
+    rabin_fingerprint,
 )
 
+SYMBOL_REGEX = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")
+
 
 def extract_record_type(schema):
     if isinstance(schema, dict):
@@ -36,12 +51,12 @@ def extract_logical_type(schema):
     return None
 
 
-def fullname(schema):
+def fullname(schema: DictSchema) -> str:
     """Returns the fullname of a schema
 
     Parameters
     ----------
-    schema: dict
+    schema
         Input schema
 
 
@@ -67,7 +82,7 @@ def fullname(schema):
     return schema_name(schema, "")[1]
 
 
-def schema_name(schema, parent_ns):
+def schema_name(schema: DictSchema, parent_ns: str) -> Tuple[str, str]:
     try:
         name = schema["name"]
     except KeyError:
@@ -84,7 +99,7 @@ def schema_name(schema, parent_ns):
         return namespace, f"{namespace}.{name}"
 
 
-def expand_schema(schema):
+def expand_schema(schema: Schema) -> Schema:
     """Returns a schema where all named types are expanded to their real schema
 
     NOTE: The output of this function produces a schema that can include
@@ -153,12 +168,13 @@ def expand_schema(schema):
 
 
 def parse_schema(
-    schema,
-    expand=False,
-    _write_hint=True,
-    _force=False,
-    _named_schemas=None,
-):
+    schema: Schema,
+    named_schemas: Optional[NamedSchemas] = None,
+    *,
+    expand: bool = False,
+    _write_hint: bool = True,
+    _force: bool = False,
+) -> Schema:
     """Returns a parsed avro schema
 
     It is not necessary to call parse_schema but doing so and saving the parsed
@@ -167,26 +183,22 @@ def parse_schema(
 
     Parameters
     ----------
-    schema: dict
+    schema
         Input schema
-    expand: bool
-        NOTE: This option should be considered a keyword only argument and may
-        get enforced as such when Python 2 support is dropped.
-
+    named_schemas
+        Dictionary of named schemas to their schema definition
+    expand
         If true, named schemas will be fully expanded to their true schemas
         rather than being represented as just the name. This format should be
         considered an output only and not passed in to other reader/writer
         functions as it does not conform to the avro specification and will
         likely cause an exception
-    _write_hint: bool
+    _write_hint
         Internal API argument specifying whether or not the __fastavro_parsed
         marker should be added to the schema
-    _force: bool
+    _force
         Internal API argument. If True, the schema will always be parsed even
         if it has been parsed and has the __fastavro_parsed marker
-    _named_schemas: dict
-        Internal API argument. Dictionary of named schemas to their schema
-        definition
 
 
     Example::
@@ -197,27 +209,61 @@ def parse_schema(
         parsed_schema = parse_schema(original_schema)
         with open('weather.avro', 'wb') as out:
             writer(out, parsed_schema, records)
+
+
+    Sometimes you might have two schemas where one schema references another.
+    For the sake of example, let's assume you have a `Parent` schema that
+    references a `Child` schema`. If you were to try to parse the parent schema
+    on its own, you would get an exception because the child schema isn't
+    defined. To accommodate this, we can use the `named_schemas` argument to pass
+    a shared dictionary when parsing both of the schemas. The dictionary will
+    get populated with the necessary schema references to make parsing possible.
+    For example::
+
+        from fastavro import parse_schema
+
+        named_schemas = {}
+        parsed_child = parse_schema(child_schema, named_schemas)
+        parsed_parent = parse_schema(parent_schema, named_schemas)
     """
-    if _named_schemas is None:
-        _named_schemas = {}
+    if named_schemas is None:
+        named_schemas = {}
+
+    if isinstance(schema, dict) and "__fastavro_parsed" in schema:
+        if "__named_schemas" in schema:
+            for key, value in schema["__named_schemas"].items():
+                named_schemas[key] = value
+        else:
+            # Some old schemas might only have __fastavro_parsed and not
+            # __named_schemas since that came later. For these schemas, we need
+            # to re-parse the schema to handle named types
+            return _parse_schema(schema, "", expand, _write_hint, set(), named_schemas)
 
     if _force or expand:
-        return _parse_schema(schema, "", expand, _write_hint, set(), _named_schemas)
+        return _parse_schema(schema, "", expand, _write_hint, set(), named_schemas)
     elif isinstance(schema, dict) and "__fastavro_parsed" in schema:
-        for key, value in schema["__named_schemas"].items():
-            _named_schemas[key] = value
         return schema
     elif isinstance(schema, list):
         # If we are given a list we should make sure that the immediate sub
         # schemas have the hint in them
         return [
-            parse_schema(s, expand, _write_hint, _force, _named_schemas) for s in schema
+            parse_schema(
+                s, named_schemas, expand=expand, _write_hint=_write_hint, _force=_force
+            )
+            for s in schema
         ]
     else:
-        return _parse_schema(schema, "", expand, _write_hint, set(), _named_schemas)
+        return _parse_schema(schema, "", expand, _write_hint, set(), named_schemas)
 
 
-def _parse_schema(schema, namespace, expand, _write_hint, names, named_schemas):
+def _parse_schema(
+    schema: Schema,
+    namespace: str,
+    expand: bool,
+    _write_hint: bool,
+    names: Set[str],
+    named_schemas: NamedSchemas,
+) -> Schema:
     # union schemas
     if isinstance(schema, list):
         return [
@@ -268,14 +314,14 @@ def _parse_schema(schema, namespace, expand, _write_hint, names, named_schemas):
             scale = parsed_schema.get("scale")
             if scale and (not isinstance(scale, int) or scale < 0):
                 raise SchemaParseException(
-                    f"decimal scale must be a postive integer, not {scale}"
+                    f"decimal scale must be a positive integer, not {scale}"
                 )
 
             precision = parsed_schema.get("precision")
             if precision:
                 if not isinstance(precision, int) or precision <= 0:
                     raise SchemaParseException(
-                        "decimal precision must be a postive integer, "
+                        "decimal precision must be a positive integer, "
                         + f"not {precision}"
                     )
                 if schema_type == "fixed":
@@ -320,6 +366,8 @@ def _parse_schema(schema, namespace, expand, _write_hint, names, named_schemas):
                 raise SchemaParseException(f"redefined named type: {fullname}")
             names.add(fullname)
 
+            _validate_enum_symbols(schema)
+
             named_schemas[fullname] = parsed_schema
 
             parsed_schema["name"] = fullname
@@ -398,32 +446,41 @@ def parse_field(field, namespace, expand, names, named_schemas):
 
 
 def load_schema(
-    schema_path, *, _named_schemas=None, _write_hint=True, _injected_schemas=None
-):
-    """Returns a schema loaded from the file at `schema_path`.
+    schema_path: str,
+    *,
+    repo: Optional[AbstractSchemaRepository] = None,
+    named_schemas: Optional[NamedSchemas] = None,
+    _write_hint: bool = True,
+    _injected_schemas: Set[str] = None,
+) -> Schema:
+    """Returns a schema loaded from repository.
 
-    Will recursively load referenced schemas assuming they can be found in
-    files in the same directory and named with the convention
-    `<full_name>.avsc`.
+    Will recursively load referenced schemas attempting to load them from
+    same repository, using `schema_path` as schema name.
+
+    If `repo` is not provided, `FlatDictRepository` is used.
+    `FlatDictRepository` will try to load schemas from the same directory
+    assuming files are named with the convention `<full_name>.avsc`.
 
     Parameters
     ----------
-    schema: str
-        Path to schema file to load
-    _named_schemas: dict
-        Internal API argument. Dictionary of named schemas to their schema
-        definition
-    _write_hint: bool
+    schema_path
+        Full schema name, or path to schema file if default repo is used.
+    repo:
+        Schema repository instance.
+    named_schemas
+        Dictionary of named schemas to their schema definition
+    _write_hint
         Internal API argument specifying whether or not the __fastavro_parsed
         marker should be added to the schema
-    _injected_schemas: set
+    _injected_schemas
         Internal API argument. Set of names that have been injected
 
 
-    Consider the following example...
+    Consider the following example with default FlatDictRepository...
 
 
-    Parent.avsc::
+    namespace.Parent.avsc::
 
         {
             "type": "record",
@@ -452,45 +509,73 @@ def load_schema(
 
         from fastavro.schema import load_schema
 
-        parsed_schema = load_schema("Parent.avsc")
+        parsed_schema = load_schema("namespace.Parent.avsc")
     """
-    if _named_schemas is None:
-        _named_schemas = {}
+    schema_name = schema_path
+    if repo is None:
+        file_dir, file_name = path.split(schema_path)
+        schema_name, _file_ext = path.splitext(file_name)
+        repo = FlatDictRepository(file_dir)
+
+    if named_schemas is None:
+        named_schemas = {}
 
     if _injected_schemas is None:
         _injected_schemas = set()
 
-    with open(schema_path) as fd:
-        schema = json.load(fd)
-    schema_dir, schema_file = path.split(schema_path)
     return _load_schema(
-        schema, schema_dir, _named_schemas, _write_hint, _injected_schemas
+        schema_name, repo, named_schemas, _write_hint, _injected_schemas
     )
 
 
-def _load_schema(schema, schema_dir, named_schemas, write_hint, injected_schemas):
+def _load_schema(schema_name, repo, named_schemas, write_hint, injected_schemas):
+    try:
+        schema = repo.load(schema_name)
+        return _parse_schema_with_repo(
+            schema,
+            repo,
+            named_schemas,
+            write_hint,
+            injected_schemas,
+        )
+    except SchemaRepositoryError as error:
+        raise error
+
+
+def _parse_schema_with_repo(
+    schema,
+    repo,
+    named_schemas,
+    write_hint,
+    injected_schemas,
+):
     try:
         schema_copy = deepcopy(named_schemas)
         return parse_schema(
-            schema, _named_schemas=named_schemas, _write_hint=write_hint
+            schema,
+            named_schemas=named_schemas,
+            _write_hint=write_hint,
         )
-    except UnknownType as e:
+    except UnknownType as error:
+        missing_subject = error.name
         try:
-            avsc = path.join(schema_dir, f"{e.name}.avsc")
-            sub_schema = load_schema(
-                avsc,
-                _named_schemas=schema_copy,
-                _write_hint=False,
-                _injected_schemas=injected_schemas,
+            sub_schema = _load_schema(
+                missing_subject,
+                repo,
+                named_schemas=schema_copy,
+                write_hint=False,
+                injected_schemas=injected_schemas,
             )
-        except IOError:
-            raise e
+        except SchemaRepositoryError:
+            raise error
 
         if sub_schema["name"] not in injected_schemas:
-            _inject_schema(schema, sub_schema)
+            injected_schema = _inject_schema(schema, sub_schema)
+            if isinstance(schema, str) or isinstance(schema, list):
+                schema = injected_schema[0]
             injected_schemas.add(sub_schema["name"])
-        return _load_schema(
-            schema, schema_dir, schema_copy, write_hint, injected_schemas
+        return _parse_schema_with_repo(
+            schema, repo, schema_copy, write_hint, injected_schemas
         )
 
 
@@ -582,3 +667,200 @@ def _inject_schema(outer_schema, inner_schema, namespace="", is_injected=False):
                 "Internal error; "
                 + "You should raise an issue in the fastavro github repository"
             )
+
+
+def load_schema_ordered(
+    ordered_schemas: List[str], *, _write_hint: bool = True
+) -> Schema:
+    """Returns a schema loaded from a list of schemas.
+
+    The list of schemas should be ordered such that any dependencies are listed
+    before any other schemas that use those dependencies. For example, if schema
+    `A` depends on schema `B` and schema B depends on schema `C`, then the list
+    of schemas should be [C, B, A].
+
+    Parameters
+    ----------
+    ordered_schemas
+        List of paths to schemas
+    _write_hint
+        Internal API argument specifying whether or not the __fastavro_parsed
+        marker should be added to the schema
+
+
+    Consider the following example...
+
+
+    Parent.avsc::
+
+        {
+            "type": "record",
+            "name": "Parent",
+            "namespace": "namespace",
+            "fields": [
+                {
+                    "name": "child",
+                    "type": "Child"
+                }
+            ]
+        }
+
+
+    namespace.Child.avsc::
+
+        {
+            "type": "record",
+            "namespace": "namespace",
+            "name": "Child",
+            "fields": []
+        }
+
+
+    Code::
+
+        from fastavro.schema import load_schema_ordered
+
+        parsed_schema = load_schema_ordered(
+            ["path/to/namespace.Child.avsc", "path/to/Parent.avsc"]
+        )
+    """
+    loaded_schemas = []
+    named_schemas: NamedSchemas = {}
+    for idx, schema_path in enumerate(ordered_schemas):
+        # _write_hint is always False except maybe the outer most schema
+        _last = _write_hint if idx + 1 == len(ordered_schemas) else False
+        schema = load_schema(
+            schema_path, named_schemas=named_schemas, _write_hint=_last
+        )
+        loaded_schemas.append(schema)
+
+    top_first_order = loaded_schemas[::-1]
+    outer_schema = top_first_order.pop(0)
+
+    while top_first_order:
+        sub_schema = top_first_order.pop(0)
+        _inject_schema(outer_schema, sub_schema)
+
+    return outer_schema
+
+
+def to_parsing_canonical_form(schema: Schema) -> str:
+    """Returns a string represening the parsing canonical form of the schema.
+
+    For more details on the parsing canonical form, see here:
+    https://avro.apache.org/docs/current/spec.html#Parsing+Canonical+Form+for+Schemas
+
+    Parameters
+    ----------
+    schema
+        Schema to transform
+
+    """
+    fo = StringIO()
+    _to_parsing_canonical_form(parse_schema(schema), fo)
+    return fo.getvalue()
+
+
+def _to_parsing_canonical_form(schema, fo):
+    # union schemas
+    if isinstance(schema, list):
+        fo.write("[")
+        for idx, s in enumerate(schema):
+            if idx != 0:
+                fo.write(",")
+            _to_parsing_canonical_form(s, fo)
+        fo.write("]")
+
+    # string schemas; this could be either a named schema or a primitive type
+    elif not isinstance(schema, dict):
+        fo.write(f'"{schema}"')
+
+    else:
+        # Remaining valid schemas must be dict types
+        schema_type = schema["type"]
+
+        if schema_type == "array":
+            fo.write(f'{{"type":"{schema_type}","items":')
+            _to_parsing_canonical_form(schema["items"], fo)
+            fo.write("}")
+
+        elif schema_type == "map":
+            fo.write(f'{{"type":"{schema_type}","values":')
+            _to_parsing_canonical_form(schema["values"], fo)
+            fo.write("}")
+
+        elif schema_type == "enum":
+            name = schema["name"]
+            fo.write(f'{{"name":"{name}","type":"{schema_type}","symbols":[')
+
+            for idx, symbol in enumerate(schema["symbols"]):
+                if idx != 0:
+                    fo.write(",")
+                fo.write(f'"{symbol}"')
+            fo.write("]}")
+
+        elif schema_type == "fixed":
+            name = schema["name"]
+            size = schema["size"]
+            fo.write(f'{{"name":"{name}","type":"{schema_type}","size":{size}}}')
+
+        elif schema_type == "record" or schema_type == "error":
+            name = schema["name"]
+            fo.write(f'{{"name":"{name}","type":"record","fields":[')
+
+            for idx, field in enumerate(schema["fields"]):
+                if idx != 0:
+                    fo.write(",")
+                name = field["name"]
+                fo.write(f'{{"name":"{name}","type":')
+                _to_parsing_canonical_form(field["type"], fo)
+                fo.write("}")
+            fo.write("]}")
+
+        elif schema_type in PRIMITIVES:
+            fo.write(f'"{schema_type}"')
+
+
+def fingerprint(parsing_canonical_form: str, algorithm: str) -> str:
+    """Returns a string represening a fingerprint/hash of the parsing canonical
+    form of a schema.
+
+    For more details on the fingerprint, see here:
+    https://avro.apache.org/docs/current/spec.html#schema_fingerprints
+
+    Parameters
+    ----------
+    parsing_canonical_form
+        The parsing canonical form of a schema
+    algorithm
+        The hashing algorithm
+
+    """
+    if algorithm not in FINGERPRINT_ALGORITHMS:
+        raise ValueError(
+            f"Unknown schema fingerprint algorithm {algorithm}. "
+            + f"Valid values include: {FINGERPRINT_ALGORITHMS}"
+        )
+
+    # Fix Java names
+    algorithm = JAVA_FINGERPRINT_MAPPING.get(algorithm, algorithm)
+
+    if algorithm == RABIN_64:
+        return rabin_fingerprint(parsing_canonical_form.encode())
+
+    h = hashlib.new(algorithm, parsing_canonical_form.encode())
+    return h.hexdigest()
+
+
+def _validate_enum_symbols(schema):
+    symbols = schema["symbols"]
+    for symbol in symbols:
+        if not isinstance(symbol, str) or not re.match(SYMBOL_REGEX, symbol):
+            raise SchemaParseException(
+                "Every symbol must match the regular expression [A-Za-z_][A-Za-z0-9_]*"
+            )
+    if len(symbols) != len(set(symbols)):
+        raise SchemaParseException("All symbols in an enum must be unique")
+
+    if "default" in schema and schema["default"] not in symbols:
+        raise SchemaParseException("Default value for enum must be in symbols list")
